"""
Utils for creating forward starting options timetable.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/eq/forward to create timetables with timestamps.
"""

import pyarrow as pa

from qablet_contracts.timetable import EVENT_SCHEMA


def forward_option_timetable(
    ccy: str,
    asset_name: str,
    strike_rate: float,
    strike_time: float,
    maturity: float,
    is_call: bool,
    track: str = "",
) -> dict:
    """Create timetable for a **Forward Starting Option**.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike_rate: the option strike in percent of fixing.
        strike_time: the time at which the strike is fixed.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = forward_option_timetable("USD", "EQ", 1.1, 1.0, 2.0, True)
        >>> tt["events"].to_pandas()
          track  time   op  quantity unit
        0   NaN   1.0  NaN       0.0  FIX
        1         2.0    >       0.0  USD
        2         2.0    +      -1.1   .K
        3         2.0    +       1.0   EQ
    """
    strike_snap = track + ".K"
    events = [
        {
            "track": None,
            "time": strike_time,
            "op": None,
            "quantity": 0,
            "unit": "FIX",  # set the strike
        },
        {
            "track": track,
            "time": maturity,
            "op": ">",
            "quantity": 0,
            "unit": ccy,
        },
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": -strike_rate if is_call else strike_rate,
            "unit": strike_snap,
        },
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": 1 if is_call else -1,
            "unit": asset_name,
        },
    ]

    # Define the strike expression, return the spot itself.
    def strike_fn(inputs):
        return inputs

    fix = {
        "type": "snapper",
        "inp": [asset_name],
        "fn": strike_fn,
        "out": [strike_snap],
    }

    events_table = pa.RecordBatch.from_pylist(events, schema=EVENT_SCHEMA)
    return {"events": events_table, "expressions": {"FIX": fix}}


if __name__ == "__main__":
    # Create the option
    timetable = forward_option_timetable("USD", "EQ", 1.1, 1.0, 2.0, True)
    print(timetable["events"].to_pandas())



from qablet_contracts.equity.autocall import autocallable_timetable
from qablet_contracts.equity.barrier import ko_option_timetable
from qablet_contracts.equity.cliquet import clique_timetable
from qablet_contracts.equity.forward_start import forward_option_timetable


def test_methods():
    tt = ko_option_timetable("USD", "EQ", 100, 0.2, True, 100, "Dn/Out", 5)
    assert len(tt["events"]) == 9

    tt = autocallable_timetable("USD", "AAPL", 100, 80, 1, 102, 4, 0.092)
    assert len(tt["events"]) == 5

    tt = clique_timetable("USD", "SPX", [1.0, 2.0, 3.0], 0.01, -0.03, 0.05)
    assert len(tt["events"]) == 5

    tt = forward_option_timetable("USD", "EQ", 1.1, 1.0, 2.0, True)
    assert len(tt["events"]) == 4



"""
This module contains examples of barrier options.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

import pandas as pd

from qablet_contracts.eq.vanilla import Option
from qablet_contracts.timetable import EventsMixin


@dataclass
class OptionKO(EventsMixin):
    """In a **Dn/Out Option** the contract is cancelled if the underlying asset price
    falls below the barrier level on any of the barrier observation dates.
    In an **Up/Out Option** the contract is cancelled if the underlying asset price
    rises above the barrier level on any of the observation dates.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike: the option strike.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        barrier: the barrier level.
        barrier_type: the type of barrier option, e.g "Dn/Out" or "Up/Out".
        barrier_dates: the barrier observation points.
        rebate: the rebate amount paid at cancellation.
        track: an optional identifier for the contract.

    Examples:
        >>> start = datetime(2024, 3, 31)
        >>> maturity = datetime(2024, 9, 30)
        >>> barrier_dates = pd.date_range(start, maturity, freq="2ME")
        >>> tt = OptionKO(
            "USD", "EQ", 100, maturity, True, 102, "Up/Out", barrier_dates
        ).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time  op  quantity unit
        0       2024-03-31 00:00:00+00:00  KO       0.0  USD
        1       2024-05-31 00:00:00+00:00  KO       0.0  USD
        2       2024-07-31 00:00:00+00:00  KO       0.0  USD
        3       2024-09-30 00:00:00+00:00  KO       0.0  USD
        4       2024-09-30 00:00:00+00:00   >       0.0  USD
        5       2024-09-30 00:00:00+00:00   +    -100.0  USD
        6       2024-09-30 00:00:00+00:00   +       1.0   EQ
    """

    ccy: str
    asset_name: str
    strike: float
    maturity: datetime
    is_call: bool
    barrier: float
    barrier_type: str
    barrier_dates: List[datetime]
    rebate: float = 0
    track: str = ""

    def events(self):
        events = []
        for barrier_date in self.barrier_dates:
            events.append(
                {
                    "track": self.track,
                    "time": barrier_date,
                    "op": "KO",
                    "quantity": self.rebate,
                    "unit": self.ccy,
                }
            )

        vanilla_events = Option(
            self.ccy,
            self.asset_name,
            self.strike,
            self.maturity,
            self.is_call,
            self.track,
        ).events()

        events.extend(vanilla_events)
        return events

    def expressions(self):
        """Define the knockout expression (KO)."""
        if self.barrier_type == "Dn/Out":

            def ko_fn(inputs):
                [S] = inputs
                return [S < self.barrier]
        elif self.barrier_type == "Up/Out":

            def ko_fn(inputs):
                [S] = inputs
                return [S > self.barrier]
        else:
            raise ValueError(f"Unknown barrier type: {self.barrier_type}")

        return {
            "KO": {
                "type": "phrase",
                "inp": [self.asset_name],
                "fn": ko_fn,
            }
        }


if __name__ == "__main__":
    # Create the KO option
    start = datetime(2024, 3, 31)
    maturity = datetime(2024, 9, 30)
    barrier_dates = pd.date_range(start, maturity, freq="2ME")
    timetable = OptionKO(
        "USD", "EQ", 100, maturity, True, 102, "Up/Out", barrier_dates
    ).timetable()
    print(timetable["events"].to_pandas())



"""
This module contains examples of interest rate swaps.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

import pandas as pd

from qablet_contracts.ir.dcf import dcf_30_360 as dcf
from qablet_contracts.timetable import EventsMixin


def simple_swap_period(
    ccy: str,
    start: datetime,
    end: datetime,
    fixed_rate: float,
    track: str = "",
) -> list:
    """Simple representation of a swap period, paying fixed, receiving floating rate.

    Args:
        ccy: the currency of the swap.
        start: the start of the period.
        end: the end of the period.
        fixed_rate: the fixed annual rate of the swap.
        track: an optional identifier for the contract.
    """
    return [
        {
            "track": track,
            "time": start,
            "op": "+",
            "quantity": 1,
            "unit": ccy,
        },
        {
            "track": track,
            "time": end,
            "op": "+",
            "quantity": -1 - fixed_rate * dcf(end, start),
            "unit": ccy,
        },
    ]


@dataclass
class Swap(EventsMixin):
    """In a **Vanilla Swap**, at the end of each period the holder pays a fixed rate and receives a floating rate.
    In this simple version the floating rate payment is replaced by receiving notional at the beginning of the period
    and paying the notional at the end of the period.

    Args:
        ccy: the currency of the swap.
        dates: the period datetimes of the swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.

    Examples:
        >>> dates = pd.bdate_range(datetime(2023, 12, 31), datetime(2024, 12, 31), freq="2QE")
        >>> tt = Swap("USD", dates, strike_rate = 0.03).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity unit
        0  .swp 2023-12-31 00:00:00+00:00  +     1.000  USD
        1  .swp 2024-06-30 00:00:00+00:00  +    -1.015  USD
        2  .swp 2024-06-30 00:00:00+00:00  +     1.000  USD
        3  .swp 2024-12-31 00:00:00+00:00  +    -1.015  USD
    """

    ccy: str
    dates: List[datetime]
    strike_rate: float
    track: str = ""

    def events(self):
        events = []
        # payment events
        for start, end in zip(self.dates[0:-1], self.dates[1:]):
            events.extend(
                simple_swap_period(
                    self.ccy, start, end, self.strike_rate, self.track + ".swp"
                )
            )

        return events


if __name__ == "__main__":
    dates = pd.bdate_range(
        datetime(2023, 12, 31),
        datetime(2025, 12, 31),
        freq="2QE",
    )
    tt = Swap("USD", dates, strike_rate=0.03).timetable()
    print(tt["events"].to_pandas())



"""
This module contains examples of equity rainbow options.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

from qablet_contracts.timetable import EventsMixin


@dataclass
class Rainbow(EventsMixin):
    """A **Rainbow Call Option** offers the holder the option to buy the best of
    a list of stocks (or none) for corresponding fixed strike prices, on the option maturity date.
    Similarly, a **Rainbow Put Option** offers the holder the option to sell the worst of
    a list of stocks (or none) for corresponding fixed strike prices.

    Args:
        ccy: the currency of the option.
        asset_names: the name of the underlying assets.
        strikes: the option strikes.
        notional: the notional of the option.
        maturity: the maturity of the option.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.

    Examples:
        >>> assets = ["SPX", "FTSE", "N225"]
        >>> strikes = [5087, 7684, 39100]
        >>> tt = Rainbow("USD", assets, strikes, 100_000, datetime(2024, 3, 31), True).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op       quantity  unit
        0       2024-03-31 00:00:00+00:00  + -100000.000000   USD
        1       2024-03-31 00:00:00+00:00  >      19.657952   SPX
        2       2024-03-31 00:00:00+00:00  >      13.014055  FTSE
        3       2024-03-31 00:00:00+00:00  >       2.557545  N225
        4       2024-03-31 00:00:00+00:00  +  100000.000000   USD
    """

    ccy: str
    asset_names: List[str]
    strikes: List[float]
    notional: float
    maturity: datetime
    is_call: bool
    track: str = ""

    def events(self):
        sign = 1 if self.is_call else -1
        # Pay the initial strike
        events = [
            {
                "track": "",
                "time": self.maturity,
                "op": "+",
                "quantity": -self.notional * sign,
                "unit": self.ccy,
            },
        ]

        # Options to receive any of the assets
        for asset, strike in zip(self.asset_names, self.strikes):
            events.append(
                {
                    "track": "",
                    "time": self.maturity,
                    "op": ">",
                    "quantity": self.notional / strike * sign,
                    "unit": asset,
                }
            )

        # Otherwise receive the notional back
        events.append(
            {
                "track": "",
                "time": self.maturity,
                "op": "+",
                "quantity": self.notional * sign,
                "unit": self.ccy,
            }
        )
        return events


if __name__ == "__main__":
    # Create the rainbow option
    timetable = Rainbow(
        "USD",
        ["SPX", "FTSE", "N225"],
        [5087, 7684, 39100],
        100_000,
        datetime(2024, 3, 31),
        True,
    ).timetable()

    print(timetable["events"].to_pandas())



"""
This module contains examples of forward starting options.
"""

from dataclasses import dataclass
from datetime import datetime

from qablet_contracts.timetable import EventsMixin


@dataclass
class ForwardOption(EventsMixin):
    """In a **Forward Starting Option** the strike price is set on a future date as
    a predetermine percent of the stock price on the strike fixing date.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike_rate: the option strike in percent of fixing.
        strike_date: the datetime at which the strike is fixed.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.

    Examples:
        >>> timetable = ForwardOption( "USD", "SPX", 1.0,
        datetime(2024, 3, 31), datetime(2024, 9, 30), True).timetable()
        >>> print(timetable["events"].to_pandas())
          track                      time   op  quantity    unit
        0   NaN 2024-03-31 00:00:00+00:00  NaN       0.0  .FIX_K
        1       2024-09-30 00:00:00+00:00    >       0.0     USD
        2       2024-09-30 00:00:00+00:00    +      -1.0      .K
        3       2024-09-30 00:00:00+00:00    +       1.0     SPX
    """

    ccy: str
    asset_name: str
    strike_rate: float
    strike_date: datetime
    maturity: datetime
    is_call: bool
    track: str = ""

    def events(self):
        sign = 1 if self.is_call else -1
        return [
            {
                "track": None,
                "time": self.strike_date,
                "op": None,
                "quantity": 0,
                "unit": f"{self.track}.FIX_K",  # set the strike
            },
            {
                "track": self.track,
                "time": self.maturity,
                "op": ">",
                "quantity": 0,
                "unit": self.ccy,
            },
            {
                "track": self.track,
                "time": self.maturity,
                "op": "+",
                "quantity": -self.strike_rate * sign,
                "unit": f"{self.track}.K",
            },
            {
                "track": self.track,
                "time": self.maturity,
                "op": "+",
                "quantity": sign,
                "unit": self.asset_name,
            },
        ]

    def expressions(self):
        # Define the strike expression, return the spot itself.
        def strike_fn(inputs):
            return inputs

        return {
            f"{self.track}.FIX_K": {
                "type": "snapper",
                "inp": [self.asset_name],
                "fn": strike_fn,
                "out": [f"{self.track}.K"],
            }
        }


if __name__ == "__main__":
    # Create the option timetable
    timetable = ForwardOption(
        "USD",
        "SPX",
        1.0,
        datetime(2024, 3, 31),
        datetime(2024, 9, 30),
        True,
    ).timetable()

    print(timetable["events"].to_pandas())



"""
This module contains examples of zero coupon bond and related contracts.
"""

from dataclasses import dataclass
from datetime import datetime

from qablet_contracts.timetable import EventsMixin


@dataclass
class Bond(EventsMixin):
    """A **zero coupon bond** pays a single fixed payment at a future time.

    Args:
        ccy: the currency of the bond.
        maturity: the maturity of the bond.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = Bond("USD", datetime(2025, 3, 31)).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity unit
        0       2025-03-31 00:00:00+00:00  +       1.0  USD
    """

    ccy: str
    maturity: datetime
    track: str = ""

    def events(self):
        return [
            {
                "track": self.track,
                "time": self.maturity,
                "op": "+",
                "quantity": 1,
                "unit": self.ccy,
            }
        ]


@dataclass
class BondPut(EventsMixin):
    """A **zero coupon bond put** offers the holder the option to sell a zero coupon bond for
    a fixed strike price, on the option maturity date.

    Args:
        ccy: the currency of the bond.
        opt_maturity: the maturity of the option.
        bond_maturity: the maturity of the bond.
        strike: the option strike.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = BondPut(
            "USD", datetime(2024, 9, 30), datetime(2025, 3, 31), 0.95
        ).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity unit
        0       2024-09-30 00:00:00+00:00  >      0.00  USD
        1       2024-09-30 00:00:00+00:00  +      0.95  USD
        2       2025-03-31 00:00:00+00:00  +     -1.00  USD
    """

    ccy: str
    opt_maturity: datetime
    bond_maturity: datetime
    strike: float
    track: str = ""

    def events(self):
        return [
            {
                "track": self.track,
                "time": self.opt_maturity,
                "op": ">",
                "quantity": 0,
                "unit": self.ccy,
            },  # Choose greater of nothing (get 0) or exercise (continue to remaining events)
            {
                "track": self.track,
                "time": self.opt_maturity,
                "op": "+",
                "quantity": self.strike,
                "unit": self.ccy,
            },  # get strike at expiration
            {
                "track": self.track,
                "time": self.bond_maturity,
                "op": "+",
                "quantity": -1,
                "unit": self.ccy,
            },  # pay bond notional at bond expiration
        ]


if __name__ == "__main__":
    # Create a zero coupon bond timetable
    timetable = Bond("USD", datetime(2025, 3, 31)).timetable()
    print("zcb:\n", timetable["events"].to_pandas())

    timetable = BondPut(
        "USD", datetime(2024, 9, 30), datetime(2025, 3, 31), 0.95
    ).timetable()
    print("zbp:\n", timetable["events"].to_pandas())



# Define the timetable schema

from abc import ABC, abstractmethod
from typing import Dict, List

import pyarrow as pa

DICT_TYPE = pa.dictionary(pa.int64(), pa.string())

# Event Schema for the timetable, using floating point for time,
# now being deprecated
EVENT_SCHEMA = pa.schema(
    [
        pa.field("track", DICT_TYPE),
        pa.field("time", pa.float64()),
        pa.field("op", DICT_TYPE),
        pa.field("quantity", pa.float64()),
        pa.field("unit", DICT_TYPE),
    ]
)

# Event Schema for the timetable, using timestamp for time
TS_TYPE = pa.timestamp("ms", tz="UTC")
TS_EVENT_SCHEMA = pa.schema(
    [
        pa.field("track", DICT_TYPE),
        pa.field("time", TS_TYPE),
        pa.field("op", DICT_TYPE),
        pa.field("quantity", pa.float64()),
        pa.field("unit", DICT_TYPE),
    ]
)


def timetable_from_dicts(events: List[Dict]) -> Dict:
    """Create timetable from a list of dicts. This method creates a timetable
    with floating point for time, which is now being deprecated.

    Args:
        events: a list of dicts with the following fields:

            - track: string
            - time: float
            - op: string
            - quantity: float
            - unit: string

    Returns:
        a timetable dictionary with the following fields:

            - events: a pyarrow record batch
            - expressions: a dict for expressions, batches, and snappers

    """
    return {
        "events": pa.RecordBatch.from_pylist(events, schema=EVENT_SCHEMA),
        "expressions": {},
    }


def py_to_ts(py_dt):
    """Convert a python datetime to a pyarrow timestamp (milliseconds)."""
    return pa.scalar(py_dt, type=TS_TYPE)


class EventsMixin(ABC):
    """A mixin class for contracts that generates a timetable from events list.
    A derived class needs to implement the events method that returns a list of dicts,
    and the expressions method (optional) that returns a dictionary of expressions, batches, and snappers."""

    @abstractmethod
    def events(self) -> List[Dict]: ...

    def expressions(self) -> Dict:
        return {}

    def timetable(self):
        return {
            "events": pa.RecordBatch.from_pylist(
                self.events(), schema=TS_EVENT_SCHEMA
            ),
            "expressions": self.expressions(),
        }


def convert_time_to_ts(timetable, base_ts):
    """Convert events of timetable in place, changing time column into timestamp column.
    This function is used convert timetables created by legacy methods that use floating point
    for time into timetables with timestamp for time."""
    events = timetable["events"]

    if events["time"].type == pa.float64():
        ts_list = [
            base_ts + t.as_py() * 31_536_000_000 for t in events["time"]
        ]

        timetable["events"] = pa.RecordBatch.from_arrays(
            [
                events["track"],
                ts_list,
                events["op"],
                events["quantity"],
                events["unit"],
            ],
            schema=TS_EVENT_SCHEMA,
        )



"""
This module contains examples of equity vanilla options.
"""

from dataclasses import dataclass
from datetime import datetime

from qablet_contracts.timetable import EventsMixin


@dataclass
class Option(EventsMixin):
    """An **European Call Option** offers the holder the option to buy a stock for
    a fixed strike price, on the option maturity date.
    Similarly, a **Put Option** offers the holder the option to sell a stock for
    a fixed strike price.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike: the option strike.
        maturity: the maturity of the option.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.

    Examples:
        >>> call = Option("USD", "SPX", 2900, datetime(2024, 3, 31), True, "<SPX2900>")
        >>> call.timetable()["events"].to_pandas()
          track  time op  quantity  unit
               track                      time op  quantity unit
        0  <SPX2900> 2024-03-31 00:00:00+00:00  >       0.0  USD
        1  <SPX2900> 2024-03-31 00:00:00+00:00  +   -2900.0  USD
        2  <SPX2900> 2024-03-31 00:00:00+00:00  +       1.0  SPX

        >>> put = Option("USD", "SPX", 2900, datetime(2024, 3, 31), False, "<SPX2900>")
        >>> put.timetable()["events"].to_pandas()
          track  time op  quantity  unit
               track                      time op  quantity unit
        0  <SPX2900> 2024-03-31 00:00:00+00:00  >       0.0  USD
        1  <SPX2900> 2024-03-31 00:00:00+00:00  +    2900.0  USD
        2  <SPX2900> 2024-03-31 00:00:00+00:00  +      -1.0  SPX
    """

    ccy: str
    asset_name: str
    strike: float
    maturity: datetime
    is_call: bool
    track: str = ""

    def events(self):
        sign = 1 if self.is_call else -1
        return [
            {
                "track": self.track,
                "time": self.maturity,
                "op": ">",
                "quantity": 0,
                "unit": self.ccy,
            },
            {
                "track": self.track,
                "time": self.maturity,
                "op": "+",
                "quantity": -self.strike * sign,
                "unit": self.ccy,
            },
            {
                "track": self.track,
                "time": self.maturity,
                "op": "+",
                "quantity": sign,
                "unit": self.asset_name,
            },
        ]


if __name__ == "__main__":
    # Create the option timetable
    timetable = Option(
        "USD",
        "SPX",
        2900,
        datetime(2024, 3, 31),
        True,
    ).timetable()

    print(timetable["events"].to_pandas())



"""
This module contains examples of autocallable notes.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

import numpy as np
import pandas as pd

from qablet_contracts.ir.dcf import dcf_30_360 as dcf
from qablet_contracts.timetable import EventsMixin


@dataclass
class DiscountCert(EventsMixin):
    """An **Autocallable Discount Certificate** is called if the asset price is above the barrier level
    on any of the barrier observation dates. If called, the note pays the principal and the coupon accreted
    till the call date. Otherwise, at maturity, if the asset is above strike, it pays the principal and the
    coupon at maturity. If the asset is below strike, the principal payment is scaled down proportionately
    with the asset price.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        initial_spot: the initial spot price of the asset.
        strike: downside participation below this strike.
        maturity: the maturity of the option in years.
        barrier: the note is called above this barrier level.
        barrier_dates: the barrier observation points.
        cpn_rate: the coupon rate.
        notional: the notional amount.
        track: an optional identifier for the contract.

    Examples:
        >>> start = datetime(2024, 3, 31)
        >>> maturity = datetime(2024, 9, 30)
        >>> barrier_dates = pd.date_range(start, maturity, freq="ME", inclusive="right")
        >>> tt = DiscountCert(
            "USD", "AAPL", 100, 80, start, maturity, 102, barrier_dates, 0.092
        ).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time    op    quantity    unit
        0       2024-04-30 00:00:00+00:00  CALL  100.769613     USD
        1       2024-05-31 00:00:00+00:00  CALL  101.545149     USD
        2       2024-06-30 00:00:00+00:00  CALL  102.326654     USD
        3       2024-07-31 00:00:00+00:00  CALL  103.114173     USD
        4       2024-08-31 00:00:00+00:00  CALL  103.907753     USD
        5       2024-09-30 00:00:00+00:00  CALL  104.707441     USD
        6       2024-09-30 00:00:00+00:00     +    1.000000  PAYOFF
    """

    ccy: str
    asset_name: str
    initial_spot: float
    strike: float
    accrual_start: datetime
    maturity: datetime
    barrier: float
    barrier_dates: List[datetime]
    cpn_rate: float
    notional: float = 100.0
    track: str = ""

    def events(self):
        events = []
        # Autocall events
        for barrier_date in self.barrier_dates:
            # daycount_fraction
            frac = dcf(barrier_date, self.accrual_start)
            events.append(
                {
                    "track": self.track,
                    "time": barrier_date,
                    "op": "CALL",
                    "quantity": self.notional * np.exp(frac * self.cpn_rate),
                    "unit": self.ccy,
                }
            )

        # payoff at maturity
        events.append(
            {
                "track": "",
                "time": self.maturity,
                "op": "+",
                "quantity": 1.0,
                "unit": "PAYOFF",
            }
        )
        return events

    def fixed_payoff(self):
        return self.notional * np.exp(
            dcf(self.maturity, self.accrual_start) * self.cpn_rate
        )

    def expressions(self):
        # Define the autocall condition
        def ko_fn(inputs):
            [S] = inputs
            return [S > (self.barrier * self.initial_spot / self.notional)]

        call = {
            "type": "phrase",
            "inp": [self.asset_name],
            "fn": ko_fn,
        }

        # Define the final payoff
        def payoff_fn(inputs):
            [s] = inputs
            eq_pay = s * (self.notional / self.initial_spot)
            return [
                np.where(eq_pay < self.strike, eq_pay, self.fixed_payoff())
            ]

        payoff = {
            "type": "phrase",
            "inp": [self.asset_name],
            "fn": payoff_fn,
        }

        return {"PAYOFF": payoff, "CALL": call}


@dataclass
class ReverseCB(DiscountCert):
    """An **Autocallable Reverse Convertible** is called if the asset price is above the barrier level
    on any of the barrier observation dates. Otherwise, at maturity, if the asset is above strike, it pays the principal and the
    coupon at maturity. If the asset is below strike, the principal payment is scaled down proportionately
    with the asset price. The note pays periodic coupons until called or matured.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        initial_spot: the initial spot price of the asset.
        strike: downside participation below this strike.
        maturity: the maturity of the option in years.
        barrier: the note is called above this barrier level.
        barrier_dates: the barrier observation points.
        cpn_rate: the coupon rate.
        notional: the notional amount.
        track: an optional identifier for the contract.

    Examples:
        >>> start = datetime(2024, 3, 31)
        >>> maturity = datetime(2024, 9, 30)
        >>> barrier_dates = pd.date_range(start, maturity, freq="ME", inclusive="right")
        >>> tt = ReverseCB(
            "USD", "AAPL", 100, 80, start, maturity, 102, barrier_dates, 0.10
        ).timetable()
        >>> print(tt["events"].to_pandas())
           track                      time    op    quantity    unit
        0        2024-04-30 00:00:00+00:00     +    0.833333     USD
        1        2024-04-30 00:00:00+00:00  CALL  100.000000     USD
        2        2024-05-31 00:00:00+00:00     +    0.833333     USD
        3        2024-05-31 00:00:00+00:00  CALL  100.000000     USD
        4        2024-06-30 00:00:00+00:00     +    0.833333     USD
        5        2024-06-30 00:00:00+00:00  CALL  100.000000     USD
        6        2024-07-31 00:00:00+00:00     +    0.833333     USD
        7        2024-07-31 00:00:00+00:00  CALL  100.000000     USD
        8        2024-08-31 00:00:00+00:00     +    0.833333     USD
        9        2024-08-31 00:00:00+00:00  CALL  100.000000     USD
        10       2024-09-30 00:00:00+00:00     +    0.833333     USD
        11       2024-09-30 00:00:00+00:00  CALL  100.000000     USD
        12       2024-09-30 00:00:00+00:00     +    1.000000  PAYOFF
    """

    def events(self):
        events = []
        cpn_start_dates = [self.accrual_start] + list(self.barrier_dates[:-1])
        # Autocall events
        for start, end in zip(cpn_start_dates, self.barrier_dates):
            # daycount_fraction
            frac = dcf(end, start)
            events.append(
                {
                    "track": self.track,
                    "time": end,
                    "op": "+",
                    "quantity": self.notional * frac * self.cpn_rate,
                    "unit": self.ccy,
                }
            )
            events.append(
                {
                    "track": self.track,
                    "time": end,
                    "op": "CALL",
                    "quantity": self.notional,
                    "unit": self.ccy,
                }
            )

        # payoff at maturity
        events.append(
            {
                "track": "",
                "time": self.maturity,
                "op": "+",
                "quantity": 1.0,
                "unit": "PAYOFF",
            }
        )
        return events

    def fixed_payoff(self):
        return self.notional


if __name__ == "__main__":
    # Create the autocallable contract
    start = datetime(2024, 3, 31)
    maturity = datetime(2024, 9, 30)
    barrier_dates = pd.date_range(
        start, maturity, freq="ME", inclusive="right"
    )
    timetable = ReverseCB(
        "USD", "AAPL", 100, 80, start, maturity, 102, barrier_dates, 0.10
    ).timetable()
    print(timetable["events"].to_pandas())



"""
This module contains examples of fixed rate bonds.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/bnd/fixed to create timetables with timestamps.
"""

from math import ceil

import numpy as np
import pyarrow as pa

from qablet_contracts.timetable import EVENT_SCHEMA


def cashflow_timetable(ccy: str, times, amounts, track: str = "") -> dict:
    """Create timetable from cashflows in a single currency. This example also shows how to create a timetable from arrays
    instead of a list of dictionaries, which is more efficient.

    Args:
        ccy: the currency of cashflows.
        times: a list or ndarray of cashflows times.
        amounts: a list or ndarray of cashflows amounts.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = cashflow_timetable("USD", [0.5, 1, 1.5], [0.05, 0.05, 1.05])
        >>> tt["events"].to_pandas()
          track  time op  quantity unit
        0         0.5  +      0.05  USD
        1         1.0  +      0.05  USD
        2         1.5  +      1.05  USD
    """
    n = len(times)
    return {
        "events": pa.RecordBatch.from_arrays(
            [
                pa.DictionaryArray.from_arrays(
                    indices=np.full(n, 0, dtype=np.int64), dictionary=[track]
                ),  # tracks
                pa.array(times),
                pa.DictionaryArray.from_arrays(
                    indices=np.full(n, 0, dtype=np.int64), dictionary=["+"]
                ),  # ops
                pa.array(amounts),
                pa.DictionaryArray.from_arrays(
                    indices=np.full(n, 0, dtype=np.int64), dictionary=[ccy]
                ),  # units
            ],
            schema=EVENT_SCHEMA,
        ),
        "expressions": {},
    }


def fixed_bond_timetable(
    ccy: str, coupon: float, maturity: float, freq: int = 2, track: str = ""
) -> dict:
    """Create timetable for a fixed rate bond.

    Args:
        ccy: the currency of cashflows.
        coupon: the coupon rate per year.
        maturity: the maturity of the bond.
        freq: the number of coupon payments per year.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = fixed_bond_timetable("USD", 0.05, 2.1)
        >>> tt["events"].to_pandas()
          track  time op  quantity unit
        0         0.1  +     0.025  USD
        1         0.6  +     0.025  USD
        2         1.1  +     0.025  USD
        3         1.6  +     0.025  USD
        4         2.1  +     1.025  USD
    """
    n = ceil(maturity * freq)  # How many payments left ?
    stub = (
        maturity * freq - n + 1
    )  # The first payment may be a stub/short coupon
    times = (np.arange(n) + stub) / freq
    amounts = np.full(n, coupon / freq)
    amounts[-1] += 1  # The last payment includes the principal
    return cashflow_timetable(ccy, times, amounts, track)


if __name__ == "__main__":
    # Create a timetable from cashflows
    timetable = cashflow_timetable("USD", [0.5, 1, 1.5], [0.05, 0.05, 1.05])
    print("cf:\n", timetable["events"].to_pandas())

    # Create a fixed bond timetable
    timetable = fixed_bond_timetable("USD", 0.05, 2.1)
    print("bond:\n", timetable["events"].to_pandas())
